@@ -1,31 +1,17 @@
 /*
  * Copyright (c) 2003-2010 Alexandre Ratchov <alex@caoua.org>
- * All rights reserved.
  *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions
- * are met:
+ * Permission to use, copy, modify, and distribute this software for any
+ * purpose with or without fee is hereby granted, provided that the above
+ * copyright notice and this permission notice appear in all copies.
  *
- * 	- Redistributions of source code must retain the above
- * 	  copyright notice, this list of conditions and the
- * 	  following disclaimer.
- *
- * 	- Redistributions in binary form must reproduce the above
- * 	  copyright notice, this list of conditions and the
- * 	  following disclaimer in the documentation and/or other
- * 	  materials provided with the distribution.
- *
- * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
- * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
- * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
- * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
- * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
- * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
- * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
- * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
- * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
- * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
- * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+ * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
+ * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
+ * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
+ * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
+ * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
+ * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
+ * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
 #include <stdio.h>
@@ -33,7 +19,8 @@
 #include <errno.h>
 #include <string.h>
 
-#include "dbg.h"
+#include "utils.h"
+#include "mididev.h"
 #include "sysex.h"
 #include "track.h"
 #include "song.h"
@@ -233,8 +220,8 @@ smf_put32(struct smf *o, unsigned *used, unsigned val)
 		(*used) += 4;
 	} else {
 		if (o->index + 4 > o->length) {
-			dbg_puts("smf_put32: bad chunk length\n");
-			dbg_panic();
+			log_puts("smf_put32: bad chunk length\n");
+			panic();
 		}
 		buf[0] = (val >> 24) & 0xff;
 		buf[1] = (val >> 16) & 0xff;
@@ -256,8 +243,8 @@ smf_put24(struct smf *o, unsigned *used, unsigned val)
 		(*used) += 3;
 	} else {
 		if (o->index + 3 > o->length) {
-			dbg_puts("smf_put24: bad chunk length\n");
-			dbg_panic();
+			log_puts("smf_put24: bad chunk length\n");
+			panic();
 		}
 		buf[0] = (val >> 16) & 0xff;
 		buf[1] = (val >> 8) & 0xff;
@@ -278,8 +265,8 @@ smf_put16(struct smf *o, unsigned *used, unsigned val)
 		(*used) += 2;
 	} else {
 		if (o->index + 2 > o->length) {
-			dbg_puts("smf_put16: bad chunk length\n");
-			dbg_panic();
+			log_puts("smf_put16: bad chunk length\n");
+			panic();
 		}
 		buf[0] = (val >> 8) & 0xff;
 		buf[1] = val & 0xff;
@@ -299,8 +286,8 @@ smf_putc(struct smf *o, unsigned *used, unsigned val)
 		(*used) ++;
 	} else {
 		if (o->index + 1 > o->length) {
-			dbg_puts("smf_putc: bad chunk length\n");
-			dbg_panic();
+			log_puts("smf_putc: bad chunk length\n");
+			panic();
 		}
 		fputc((int)val & 0xff, o->file);
 		o->index++;
@@ -328,8 +315,8 @@ smf_putvar(struct smf *o, unsigned *used, unsigned val)
 				(*used) += index;
 			} else {
 				if (o->index + index > o->length) {
-					dbg_puts("smf_putvar: bad chunk length\n");
-					dbg_panic();
+					log_puts("smf_putvar: bad chunk length\n");
+					panic();
 				}
 				o->index += index;
 				fwrite(buf, 1, index, o->file);
@@ -337,8 +324,8 @@ smf_putvar(struct smf *o, unsigned *used, unsigned val)
 			return;
 		}
 	}
-	dbg_puts("smf_putvar: bogus value\n");
-	dbg_panic();
+	log_puts("smf_putvar: bogus value\n");
+	panic();
 #undef MAXBYTES
 }
 
@@ -378,7 +365,8 @@ smf_puttrack(struct smf *o, unsigned *used, struct song *s, struct track *t)
 			break;
 		}
 		if (EV_ISVOICE(&pos->ev)) {
-			nev = conv_unpackev(&slist, ~1U, &pos->ev, rev);
+			nev = conv_unpackev(&slist, 0U,
+			    CONV_XPC | CONV_NRPN | CONV_RPN, &pos->ev, rev);
 			for (i = 0; i < nev; i++) {
 				smf_putvar(o, used, delta);
 				delta = 0;
@@ -424,8 +412,8 @@ smf_puttrack(struct smf *o, unsigned *used, struct song *s, struct track *t)
 				denom = 4;
 				break;
 			default:
-				dbg_puts("smf_puttrack: bad time signature\n");
-				dbg_panic();
+				log_puts("smf_puttrack: bad time signature\n");
+				panic();
 			}
 			smf_putvar(o, used, delta);
 			delta = 0;
@@ -512,7 +500,9 @@ song_exportsmf(struct song *o, char *filename)
 		ntrks++;
 	}
 	nchan = 0;
-	SONG_FOREACH_OUT(o, i) {
+	SONG_FOREACH_CHAN(o, i) {
+		if (i->isinput)
+			continue;
 		nchan++;
 	}
 	nsx = 0;
@@ -549,7 +539,9 @@ song_exportsmf(struct song *o, char *filename)
 	/*
 	 * write each chan
 	 */
-	SONG_FOREACH_OUT(o, i) {
+	SONG_FOREACH_CHAN(o, i) {
+		if (i->isinput)
+			continue;
 		used = 0;
 		smf_puttrack(&f, &used, o, &i->conf);
 		smf_putheader(&f, smftype_track, used);
@@ -601,7 +593,9 @@ smf_gettrack(struct smf *o, struct song *s, struct songtrk *t)
 	struct songsx *songsx;
 	struct seqev *pos, *se;
 	struct sysex *sx;
+	struct mididev *dev;
 	struct ev ev, rev;
+	unsigned xctlset, evset;
 	unsigned c;
 
 	if (!smf_getheader(o, smftype_track)) {
@@ -612,6 +606,9 @@ smf_gettrack(struct smf *o, struct song *s, struct songtrk *t)
 	track_clear(&t->track);
 	pos = t->track.first;
 	songsx = (struct songsx *)s->sxlist;	/* first (and unique) sysex in song */
+	if (songsx == NULL) {
+		songsx = song_sxnew(s, "smf");
+	}
 	statelist_init(&slist);
 	for (;;) {
 		if (o->index >= o->length) {
@@ -661,16 +658,16 @@ smf_gettrack(struct smf *o, struct song *s, struct songtrk *t)
 					goto err;
 				}
 				/*
-				dbg_puts("cc=");
-				dbg_putu(dummy);
-				dbg_puts(", dd=");
+				log_puts("cc=");
+				log_putu(dummy);
+				log_puts(", dd=");
 				*/
 				if (!smf_getc(o, &dummy)) {
 					goto err;
 				}
 				/*
-				dbg_putu(dummy);
-				dbg_puts("\n");
+				log_putu(dummy);
+				log_puts("\n");
 				*/
 				goto putev;
 			} else {
@@ -707,8 +704,8 @@ smf_gettrack(struct smf *o, struct song *s, struct songtrk *t)
 			} else {
 				cons_err("corrupted sysex message, ignored");
 				/*
-				sysex_dbg(sx);
-				dbg_puts("\n");
+				sysex_log(sx);
+				log_puts("\n");
 				*/
 				sysex_del(sx);
 			}
@@ -738,17 +735,35 @@ smf_gettrack(struct smf *o, struct song *s, struct songtrk *t)
 				ev.cmd = EV_NOFF;
 				ev.note_vel = EV_NOFF_DEFAULTVEL;
 			}
-			if (conv_packev(&slist, 0U, &ev, &rev)) {
+
+			/*
+			 * Pack events according to device setup.
+			 *
+			 * XXX: this needs to be done in
+			 * doevset/doxctl by converting the whole
+			 * project whenever events configuration
+			 * is changed.
+			 */
+			if ((evinfo[ev.cmd].flags & EV_HAS_DEV) &&
+			    (dev = mididev_byunit[ev.dev]) != NULL) {
+				xctlset = dev->oxctlset;
+				evset = dev->oevset;
+			} else {
+				xctlset = 0;
+				evset = CONV_XPC | CONV_NRPN | CONV_RPN;
+			}
+			if (conv_packev(&slist, xctlset, evset,
+				&ev, &rev)) {
 				se = seqev_new();
 				se->ev = rev;
 				seqev_ins(pos, se);
 			}
 			/*
-			dbg_puts("ev: ");
-			dbg_putu(abspos);
-			dbg_puts(" ");
-			ev_dbg(&ev);
-			dbg_puts("\n");
+			log_puts("ev: ");
+			log_putu(abspos);
+			log_puts(" ");
+			ev_log(&ev);
+			log_puts("\n");
 			*/
 		} else if (c < 0x80) {
 			if (status == 0) {
@@ -888,7 +903,6 @@ song_importsmf(char *filename)
 	struct songtrk *t;
 	unsigned format, ntrks, timecode, i;
 	char trackname[MAXTRACKNAME];
-	struct songsx *sx;
 	struct smf f;
 
 	if (!smf_open(&f, filename, "r")) {
@@ -907,6 +921,10 @@ song_importsmf(char *filename)
 	if (!smf_get16(&f, &ntrks)) {
 		goto bad2;
 	}
+	if (ntrks >= 256) {
+		cons_err("too many tracks in midi file");
+		goto bad2;
+	}
 	if (!smf_get16(&f, &timecode)) {
 		goto bad2;
 	}
@@ -916,8 +934,7 @@ song_importsmf(char *filename)
 	}
 
 	o = song_new();
-	song_setunit(o, timecode * 4);   /* timecode = tics per quarter */
-	sx = song_sxnew(o, "sx");
+	o->tics_per_unit = timecode * 4;   /* timecode = tics per quarter */
 
 	for (i = 0; i < ntrks; i++) {
 		snprintf(trackname, MAXTRACKNAME, "trk%02u", i);
@@ -945,3 +962,75 @@ bad3:	song_delete(o);
 bad2:	smf_close(&f);
 bad1:	return 0;
 }
+
+int
+syx_import(char *path, struct sysexlist *l, int unit)
+{
+	FILE *f;
+	struct sysex *sx = NULL;
+	int c;
+
+	f = fopen(path, "r");
+	if (f == NULL) {
+		cons_errs(path, "failed to open file");
+		return 0;
+	}
+	sysexlist_clear(l);
+	for (;;) {
+		c = fgetc(f);
+		if (c == EOF)
+			break;
+		if (c == 0xf0)
+			sx = sysex_new(unit);
+		if (sx == NULL) {
+			cons_errs(path, "corrupted .syx file");
+			goto err;
+		}
+		sysex_add(sx, c);
+		if (c == 0xf7) {
+			if (sysex_check(sx)) {
+				sysexlist_put(l, sx);
+				sx = NULL;
+				continue;
+			} else {
+				cons_err("corrupted sysex message, ignored");
+				sysex_del(sx);
+				goto err;
+			}
+		}
+	}
+	fclose(f);
+	return 1;
+err:
+	sysexlist_clear(l);
+	fclose(f);
+	return 0;
+}
+
+int
+syx_export(char *path, struct sysexlist *l)
+{
+	FILE *f;
+	struct sysex *x;
+	struct chunk *c;
+	ssize_t n;
+
+	f = fopen(path, "w");
+	if (f == NULL) {
+		cons_errs(path, "failed to open file");
+		return 0;
+	}
+	for (x = l->first; x != NULL; x = x->next) {
+		for (c = x->first; c != NULL; c = c->next) {
+			n = fwrite(c->data, 1, c->used, f);
+			if (n != c->used) {
+				cons_errs(path, "write failed");
+				fclose(f);
+				return 0;
+			}
+
+		}
+	}
+	fclose(f);
+	return 1;
+}
