@@ -1,34 +1,20 @@
 /*
- * Copyright (c) 2003-2010 Alexandre Ratchov <alex@caoua.org>
- * All rights reserved.
+ * Copyright (c) 2003-2019 Alexandre Ratchov <alex@caoua.org>
  *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions
- * are met:
+ * Permission to use, copy, modify, and distribute this software for any
+ * purpose with or without fee is hereby granted, provided that the above
+ * copyright notice and this permission notice appear in all copies.
  *
- * 	- Redistributions of source code must retain the above
- * 	  copyright notice, this list of conditions and the
- * 	  following disclaimer.
- *
- * 	- Redistributions in binary form must reproduce the above
- * 	  copyright notice, this list of conditions and the
- * 	  following disclaimer in the documentation and/or other
- * 	  materials provided with the distribution.
- *
- * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
- * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
- * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
- * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
- * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
- * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
- * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
- * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
- * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
- * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
- * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+ * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
+ * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
+ * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
+ * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
+ * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
+ * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
+ * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
-#include "dbg.h"
+#include "utils.h"
 #include "mididev.h"
 #include "mux.h"
 #include "track.h"
@@ -40,8 +26,14 @@
 #include "defs.h"
 #include "mixout.h"
 #include "norm.h"
+#include "undo.h"
+
+#define TAG_OFF		0
+#define TAG_PLAY	1
+#define TAG_REC		2
 
 unsigned song_debug = 0;
+char *song_tap_modestr[3] = {"off", "start", "tempo"};
 
 /*
  * allocate and initialize a song structure
@@ -50,7 +42,7 @@ struct song *
 song_new(void)
 {
 	struct song *o;
-	o = mem_alloc(sizeof(struct song), "song");
+	o = xmalloc(sizeof(struct song), "song");
 	song_init(o);
 	return o;
 }
@@ -62,7 +54,7 @@ void
 song_delete(struct song *o)
 {
 	song_done(o);
-	mem_free(o);
+	xfree(o);
 }
 
 /*
@@ -78,14 +70,16 @@ song_init(struct song *o)
 	 */
 	o->mode = 0;
 	o->trklist = NULL;
-	o->outlist = NULL;
-	o->inlist = NULL;
+	o->chanlist = NULL;
 	o->filtlist = NULL;
 	o->sxlist = NULL;
+	o->undo = NULL;
+	o->undo_size = 0;
 	o->tics_per_unit = DEFAULT_TPU;
 	track_init(&o->meta);
 	track_init(&o->clip);
 	track_init(&o->rec);
+	sysexlist_init(&o->recsx);
 
 	/*
 	 * runtime play record parameters
@@ -100,6 +94,7 @@ song_init(struct song *o)
 	 */
 	metro_init(&o->metro);
 	o->tic = o->beat = o->measure = 0;
+	o->abspos = 0;
 
 	/*
 	 * defaults
@@ -112,7 +107,11 @@ song_init(struct song *o)
 	o->curpos = 0;
 	o->curlen = 0;
 	o->curquant = 0;
+	o->loop = 0;
 	evspec_reset(&o->curev);
+	evspec_reset(&o->tap_evspec);
+	o->tap_evspec.cmd = EVSPEC_EMPTY;
+	o->tap_mode = 0;
 
 	/*
 	 * add default timesig/tempo so that setunit() works
@@ -137,14 +136,12 @@ song_done(struct song *o)
 	if (mux_isopen) {
 		song_stop(o);
 	}
+	undo_clear(o, &o->undo);
 	while (o->trklist) {
 		song_trkdel(o, (struct songtrk *)o->trklist);
 	}
-	while (o->outlist) {
-		song_chandel(o, (struct songchan *)o->outlist, 0);
-	}
-	while (o->inlist) {
-		song_chandel(o, (struct songchan *)o->inlist, 1);
+	while (o->chanlist) {
+		song_chandel(o, (struct songchan *)o->chanlist);
 	}
 	while (o->filtlist) {
 		song_filtdel(o, (struct songfilt *)o->filtlist);
@@ -155,7 +152,12 @@ song_done(struct song *o)
 	track_done(&o->meta);
 	track_done(&o->clip);
 	track_done(&o->rec);
+	sysexlist_done(&o->recsx);
 	metro_done(&o->metro);
+	if (o->undo != NULL) {
+		log_puts("undo data not freed\n");
+		panic();
+	}
 }
 
 /*
@@ -166,7 +168,7 @@ song_trknew(struct song *o, char *name)
 {
 	struct songtrk *t;
 
-	t = mem_alloc(sizeof(struct songtrk), "songtrk");
+	t = xmalloc(sizeof(struct songtrk), "songtrk");
 	name_init(&t->name, name);
 	track_init(&t->track);
 	t->curfilt = NULL;
@@ -190,7 +192,7 @@ song_trkdel(struct song *o, struct songtrk *t)
 	name_remove(&o->trklist, (struct name *)t);
 	track_done(&t->track);
 	name_done(&t->name);
-	mem_free(t);
+	xfree(t);
 }
 
 /*
@@ -209,22 +211,34 @@ struct songchan *
 song_channew(struct song *o, char *name, unsigned dev, unsigned ch, int input)
 {
 	struct songfilt *f;
-	struct songchan *c;
-	struct name **list = input ? &o->inlist : &o->outlist;
+	struct songchan *c, *i;
+	struct evspec src, dst;
 
-	c = mem_alloc(sizeof(struct songchan), "songchan");
+	c = xmalloc(sizeof(struct songchan), "songchan");
 	name_init(&c->name, name);
 	track_init(&c->conf);
-	c->link = NULL;
 	c->dev = dev;
 	c->ch = ch;
-	name_add(list, (struct name *)c);
-	if (!input) {
+	c->isinput = input;
+	name_add(&o->chanlist, (struct name *)c);
+	if (input)
+		c->filt = NULL;
+	else {
 		f = song_filtlookup(o, name);
 		if (f == NULL)
 			f = song_filtnew(o, name);
-		f->link = c;
-		c->link = f;
+		c->filt = f;
+		evspec_reset(&src);
+		evspec_reset(&dst);
+		dst.dev_min = dst.dev_max = c->dev;
+		dst.ch_min = dst.ch_max = c->ch;
+		SONG_FOREACH_CHAN(o, i) {
+			if (!i->isinput)
+				continue;
+			src.dev_min = src.dev_max = i->dev;
+			src.ch_min = src.ch_max = i->ch;
+			filt_mapnew(&c->filt->filt, &src, &dst);
+		}
 	}
 	song_setcurchan(o, c, input);
 	return c;
@@ -234,28 +248,21 @@ song_channew(struct song *o, char *name, unsigned dev, unsigned ch, int input)
  * delete the given chan from the song
  */
 void
-song_chandel(struct song *o, struct songchan *c, int input)
+song_chandel(struct song *o, struct songchan *c)
 {
-	struct songfilt *f;
-	struct name **list = input ? &o->inlist : &o->outlist;
-	
-	if (input) {
+	if (c->isinput) {
 		if (o->curin == c)
 			o->curin = NULL;
 	} else {
 		if (o->curout == c)
 			o->curout = NULL;
 	}
-	name_remove(list, (struct name *)c);
+	name_remove(&o->chanlist, (struct name *)c);
 	track_done(&c->conf);
 	name_done(&c->name);
-	f = c->link;
-	if (f != NULL) {
-		f->link = NULL;
-		c->link = NULL;
-		song_filtdel(o, f);
-	}
-	mem_free(c);
+	if (c->filt != NULL)
+		song_filtdel(o, c->filt);
+	xfree(c);
 }
 
 /*
@@ -264,9 +271,14 @@ song_chandel(struct song *o, struct songchan *c, int input)
 struct songchan *
 song_chanlookup(struct song *o, char *name, int input)
 {
-	struct name **list = input ? &o->inlist : &o->outlist;
+	struct songchan *c;
 
-	return (struct songchan *)name_lookup(list, name);
+	SONG_FOREACH_CHAN(o, c) {
+		if (!!c->isinput == !!input && str_eq(c->name.str, name))
+			break;
+	}
+
+	return c;
 }
 
 /*
@@ -275,15 +287,14 @@ song_chanlookup(struct song *o, char *name, int input)
 struct songchan *
 song_chanlookup_bynum(struct song *o, unsigned dev, unsigned ch, int input)
 {
-	struct songchan *i;
-	struct name *list = input ? o->inlist : o->outlist;
+	struct songchan *c;
 
-	SONG_FOREACH_CHAN(o, i, list) {
-		if (i->dev == dev && i->ch == ch) {
-			return i;
-		}
+	SONG_FOREACH_CHAN(o, c) {
+		if (!!c->isinput == !!input && c->dev == dev && c->ch == ch)
+			break;
 	}
-	return 0;
+
+	return c;
 }
 
 /*
@@ -294,10 +305,9 @@ song_filtnew(struct song *o, char *name)
 {
 	struct songfilt *f;
 
-	f = mem_alloc(sizeof(struct songfilt), "songfilt");
+	f = xmalloc(sizeof(struct songfilt), "songfilt");
 	name_init(&f->name, name);
 	filt_init(&f->filt);
-	f->link = NULL;
 	name_add(&o->filtlist, (struct name *)f);
 	song_setcurfilt(o, f);
 	return f;
@@ -309,7 +319,6 @@ song_filtnew(struct song *o, char *name)
 void
 song_filtdel(struct song *o, struct songfilt *f)
 {
-	struct songchan *c;
 	struct songtrk *t;
 
 	if (o->curfilt == f) {
@@ -323,13 +332,7 @@ song_filtdel(struct song *o, struct songfilt *f)
 	name_remove(&o->filtlist, (struct name *)f);
 	filt_done(&f->filt);
 	name_done(&f->name);
-	c = f->link;
-	if (c != NULL) {
-		c->link = NULL;
-		f->link = NULL;
-		song_chandel(o, c, 0);
-	}
-	mem_free(f);
+	xfree(f);
 }
 
 /*
@@ -349,7 +352,7 @@ song_sxnew(struct song *o, char *name)
 {
 	struct songsx *x;
 
-	x = mem_alloc(sizeof(struct songsx), "songsx");
+	x = xmalloc(sizeof(struct songsx), "songsx");
 	name_init(&x->name, name);
 	sysexlist_init(&x->sx);
 	name_add(&o->sxlist, (struct name *)x);
@@ -369,7 +372,7 @@ song_sxdel(struct song *o, struct songsx *x)
 	name_remove(&o->sxlist, (struct name *)x);
 	sysexlist_done(&x->sx);
 	name_done(&x->name);
-	mem_free(x);
+	xfree(x);
 }
 
 /*
@@ -421,13 +424,13 @@ song_setcurfilt(struct song *o, struct songfilt *f)
 {
 	if (o->curfilt == f)
 		return;
-
+	if (mux_isopen)
+		norm_shut();
 	o->curfilt = f;
-	if (f != NULL)
-		song_setcurchan(o, f->link, 0);
-	if (mux_isopen) {
-		norm_setfilt(f != NULL ? &f->filt : NULL);
-	}
+	if (o->curout && o->curout->filt != f)
+		o->curout = NULL;
+	if (mux_isopen)
+		mux_flush();
 }
 
 void
@@ -445,25 +448,7 @@ song_setcurchan(struct song *o, struct songchan *c, int input)
 		return;
 	*pc = c;
 	if (c != NULL && !input)
-		song_setcurfilt(o, c->link);
-}
-
-void
-song_setunit(struct song *o, unsigned tpu)
-{
-	struct songtrk *t;
-	unsigned mult;
-
-	mult = (tpu + DEFAULT_TPU - 1) / DEFAULT_TPU;
-	if (mult == 0)
-		mult = 1;
-	tpu = mult * DEFAULT_TPU;
-	SONG_FOREACH_TRK(o, t) {
-		track_scale(&t->track, o->tics_per_unit, tpu);
-	}
-	track_scale(&o->meta, o->tics_per_unit, tpu);
-	o->curquant = o->curquant * tpu / o->tics_per_unit;
-	o->tics_per_unit = tpu;
+		song_setcurfilt(o, c->filt);
 }
 
 unsigned
@@ -501,25 +486,25 @@ song_endpos(struct song *o)
 }
 
 void
-song_playconfev(struct song *o, struct songchan *c, int input, struct ev *in)
+song_playconfev(struct song *o, struct songchan *c, struct ev *in)
 {
 	struct ev ev = *in;
 
 	if (!EV_ISVOICE(&ev)) {
-		dbg_puts("song_playconfev: ");
-		dbg_puts(c->name.str);
-		dbg_puts(": ");
-		ev_dbg(&ev);
-		dbg_puts(": not a voice event");
-		dbg_puts("\n");
+		log_puts("song_playconfev: ");
+		log_puts(c->name.str);
+		log_puts(": ");
+		ev_log(&ev);
+		log_puts(": not a voice event");
+		log_puts("\n");
 		return;
 	}
 	ev.dev = c->dev;
 	ev.ch = c->ch;
-	if (!input) {
-		mixout_putev(&ev, 0);
+	if (!c->isinput) {
+		mixout_putev(&ev, PRIO_CHAN);
 	} else {
-		norm_putev(&ev);
+		norm_evcb(&ev);
 	}
 }
 
@@ -533,23 +518,13 @@ song_playconf(struct song *o)
 	struct seqptr *cp;
 	struct state *st;
 
-	SONG_FOREACH_IN(o, i) {
+	SONG_FOREACH_CHAN(o, i) {
 		cp = seqptr_new(&i->conf);
 		for (;;) {
 			st = seqptr_evget(cp);
 			if (st == NULL)
 				break;
-			song_playconfev(o, i, 1, &st->ev);
-		}
-		seqptr_del(cp);
-	}
-	SONG_FOREACH_OUT(o, i) {
-		cp = seqptr_new(&i->conf);
-		for (;;) {
-			st = seqptr_evget(cp);
-			if (st == NULL)
-				break;
-			song_playconfev(o, i, 0, &st->ev);
+			song_playconfev(o, i, &st->ev);
 		}
 		seqptr_del(cp);
 	}
@@ -588,7 +563,7 @@ song_metaput(struct song *o, struct state *s)
 	case EV_TIMESIG:
 		if ((s->flags & STATE_CHANGED) && (o->beat != 0 || o->tic != 0)) {
 			/*
-			 * found an incomplete measure, 
+			 * found an incomplete measure,
 			 * skip to the beggining of the next one
 			 */
 			o->beat = 0;
@@ -608,6 +583,196 @@ song_metaput(struct song *o, struct state *s)
 	}
 }
 
+/*
+ * save the state at the given start position, so that we can repeat
+ * playback from there.
+ */
+void
+song_loop_init(struct song *o)
+{
+	struct state *s, *snext;
+	struct statelist *slist;
+	struct songtrk *t;
+	unsigned qstep;
+
+	if (o->loop) {
+		o->loop_mstart = o->curpos;
+		o->loop_mend = o->curpos + o->curlen;
+	} else
+		o->loop_mstart = o->loop_mend = 0;
+
+	if (o->loop_mstart == o->loop_mend)
+		return;
+
+	o->loop_tstart = track_findmeasure(&o->meta, o->loop_mstart);
+	o->loop_tend = track_findmeasure(&o->meta, o->loop_mend);
+
+	qstep = o->curquant / 2;
+	if (o->loop_tstart > qstep) {
+		o->loop_tstart -= qstep;
+		o->loop_tend -= qstep;
+	}
+
+	o->loop_metaptr = seqptr_new(&o->meta);
+	seqptr_skip(o->loop_metaptr, o->loop_tstart);
+
+	SONG_FOREACH_TRK(o, t) {
+		t->loop_trackptr = seqptr_new(&t->track);
+		seqptr_skip(t->loop_trackptr, o->loop_tstart);
+
+		/*
+		 * Drop notes, as we don't restore them
+		 */
+		slist = &t->loop_trackptr->statelist;
+		for (s = slist->first; s != NULL; s = snext) {
+			snext = s->next;
+			if (EV_ISNOTE(&s->ev)) {
+				statelist_rm(slist, s);
+				state_del(s);
+			}
+		}
+
+		/*
+		 * Drop terminated states
+		 */
+		statelist_outdate(slist);
+	}
+}
+
+/*
+ * free loop state
+ */
+void
+song_loop_done(struct song *o)
+{
+	struct songtrk *t;
+
+	if (o->loop_mstart == o->loop_mend)
+		return;
+
+	seqptr_del(o->loop_metaptr);
+	SONG_FOREACH_TRK(o, t) {
+		statelist_empty(&t->loop_trackptr->statelist);
+		seqptr_del(t->loop_trackptr);
+	}
+}
+
+/*
+ * restore the given track (or meta-track if NULL) from its loop
+ * state.
+ */
+void
+song_loop_track(struct song *o, struct songtrk *t)
+{
+	struct seqptr *sp, *lp;
+	struct statelist *dlist, *slist;
+	struct state *s, *d, *dnext;
+	struct ev re;
+
+	if (t) {
+		sp = t->trackptr;
+		lp = t->loop_trackptr;
+	} else {
+		sp = o->metaptr;
+		lp = o->loop_metaptr;
+	}
+
+	dlist = &sp->statelist;
+	slist = &lp->statelist;
+
+	/*
+	 * cancel states not present in loop state (this will cancel
+	 * all notes as their state is not saved)
+	 */
+	for (d = dlist->first; d != NULL; d = dnext) {
+		dnext = d->next;
+		s = statelist_lookup(slist, &d->ev);
+		if (s != NULL)
+			continue;
+		if (!state_cancel(d, &re))
+			continue;
+		statelist_update(dlist, &re);
+		if (d->tag) {
+			if (t != NULL)
+				mixout_putev(&d->ev, PRIO_TRACK);
+			else
+				song_metaput(o, d);
+		}
+	}
+
+	/*
+	 * restore states from loop start
+	 */
+	for (s = slist->first; s != NULL; s = s->next) {
+		d = statelist_lookup(dlist, &s->ev);
+		if (d != NULL && state_eq(d, &s->ev))
+			continue;
+		if (!state_restore(s, &re))
+			continue;
+		d = statelist_update(dlist, &re);
+		if (d->phase & EV_PHASE_FIRST) {
+			d->tag = (t != NULL) ?
+			    !t->mute : EV_ISMETA(&d->ev);
+		}
+		if (d->tag) {
+			if (t != NULL)
+				mixout_putev(&d->ev, PRIO_TRACK);
+			else
+				song_metaput(o, d);
+		}
+	}
+
+	sp->pos = lp->pos;
+	sp->delta = lp->delta;
+	sp->tic = lp->tic;
+}
+
+void
+song_loop_rec(struct song *o)
+{
+	if (o->playptr != NULL)
+		return;
+
+	o->playptr = seqptr_new(&o->rec);
+	seqptr_link(o->playptr, o->recptr);
+
+	if (seqptr_ticdel(o->playptr, o->loop_tstart,
+		&o->rec_replay) != o->loop_tstart) {
+		log_puts("song_loop_rec: events in the way\n");
+		panic();
+	}
+	seqptr_ticput(o->playptr, o->loop_tstart);
+
+	if (song_debug)
+		log_puts("song_loop_rec: starting replay\n");
+}
+
+/*
+ * continue playback from the loop start position
+ */
+int
+song_loop_repeat(struct song *o)
+{
+	struct songtrk *t;
+
+	if (o->loop_mstart == o->loop_mend || o->abspos != o->loop_tend)
+		return 0;
+
+	o->abspos = o->loop_tstart;
+	o->measure -= o->loop_mend - o->loop_mstart;
+
+	SONG_FOREACH_TRK(o, t) {
+		song_loop_track(o, t);
+	}
+
+	song_loop_track(o, NULL);
+
+	if (o->mode >= SONG_REC)
+		song_loop_rec(o);
+
+	return 1;
+}
+
 /*
  * move all track pointers 1 tick forward. Return true if at least one
  * track moved forward and 0 if no track moved forward (ie end of the
@@ -619,8 +784,11 @@ song_metaput(struct song *o, struct state *s)
 void
 song_ticskip(struct song *o)
 {
+	struct ev ev;
 	struct songtrk *i;
+	struct state *s;
 	unsigned neot;
+	unsigned period;
 
 	/*
 	 * tempo_track
@@ -635,22 +803,42 @@ song_ticskip(struct song *o)
 			o->measure++;
 		}
 	}
-#if 0
-	if (song_debug) {
-		dbg_puts("song_ticskip: ");
-		dbg_putu(o->measure);
-		dbg_puts(":");
-		dbg_putu(o->beat);
-		dbg_puts(":");
-		dbg_putu(o->tic);
-		dbg_puts("\n");
-	}
-#endif
+	o->abspos++;
 	SONG_FOREACH_TRK(o, i) {
 		neot |= seqptr_ticskip(i->trackptr, 1);
 	}
-	if (neot == 0)
+	if (o->mode >= SONG_REC) {
+		if (o->playptr) {
+			seqptr_ticdel(o->playptr, 1, &o->rec_replay);
+			seqptr_ticput(o->playptr, 1);
+		}
+		seqptr_ticput(o->recptr, 1);
+		statelist_outdate(&o->rec_input);
+
+		/*
+		 * increment length and terminate notes longer than
+		 * the loop period
+		 */
+		period = o->loop_tend - o->loop_tstart;
+		for (s = o->rec_input.first; s != NULL; s = s->next) {
+			if (s->tag != TAG_REC)
+				continue;
+			if (++s->tic != period)
+				continue;
+			if (state_cancel(s, &ev)) {
+				seqptr_evmerge2(o->recptr,
+				    &o->rec_replay, &ev, NULL);
+				mixout_putev(&ev, 0);
+			}
+			s->tag = TAG_OFF;
+		}
+	}
+	if (song_loop_repeat(o))
+		return;
+	if (neot == 0 && !o->complete) {
+		cons_puttag("complete");
 		o->complete = 1;
+	}
 }
 
 /*
@@ -661,12 +849,12 @@ void
 song_ticplay(struct song *o)
 {
 	struct songtrk *i;
-	struct state *st;
+	struct state *st, *sr;
 	unsigned id;
 
-	while ((st = seqptr_evget(o->metaptr))) {
+	while ((st = seqptr_evget(o->metaptr)))
 		song_metaput(o, st);
-	}
+
 	if (o->tic == 0) {
 		cons_putpos(o->measure, o->beat, o->tic);
 	}
@@ -677,9 +865,30 @@ song_ticplay(struct song *o)
 			if (st->phase & EV_PHASE_FIRST)
 				st->tag = i->mute ? 0 : 1;
 			if (st->tag)
-				mixout_putev(&st->ev, id);
+				mixout_putev(&st->ev, PRIO_TRACK);
 		}
 	}
+
+	if (o->mode >= SONG_REC) {
+		/*
+		 * remove all events from rec track and put them back
+		 * on it by merging them with the input states. The
+		 * played states table is updated so it always contain
+		 * the exact state of the original track.
+		 */
+		if (o->playptr != NULL) {
+			for(;;) {
+				st = seqptr_evdel(o->playptr, &o->rec_replay);
+				if (st == NULL)
+					break;
+				st->tag = 1;
+				sr = seqptr_evmerge1(o->recptr, st);
+				if (sr != NULL)
+					mixout_putev(&st->ev, 0);
+			}
+		}
+
+	}
 }
 
 /*
@@ -687,31 +896,41 @@ song_ticplay(struct song *o)
  * that have context, then the rest.
  */
 void
-song_confrestore(struct statelist *slist)
+song_confrestore(struct statelist *slist, int all, unsigned prio)
 {
 	struct state *s;
 	struct ev re;
 
 	for (s = slist->first; s != NULL; s = s->next) {
-		if (!s->tag && !EV_ISNOTE(&s->ev)) {
-			if (state_restore(s, &re)) {
-				if (song_debug) {
-					dbg_puts("song_strestore: ");
-					ev_dbg(&s->ev);
-					dbg_puts(": restored -> ");
-					ev_dbg(&re);
-					dbg_puts("\n");
-				}
-				mixout_putev(&re, slist->serial);
-			}
-			s->tag = 1;
-		} else {
+		if (EV_ISNOTE(&s->ev))
+			continue;
+		if (s->tag) {
 			if (song_debug) {
-				dbg_puts("song_strestore: ");
-				ev_dbg(&s->ev);
-				dbg_puts(": not restored (not tagged)\n");
+				log_puts("song_strestore: ");
+				ev_log(&s->ev);
+				log_puts(": not restored (tagged)\n");
 			}
+			continue;
 		}
+		if (!(s->phase & EV_PHASE_LAST) && !all) {
+			if (song_debug) {
+				log_puts("song_strestore: ");
+				ev_log(&s->ev);
+				log_puts(": not restored (unterminated)\n");
+			}
+			continue;
+		}
+		if (state_restore(s, &re)) {
+			if (song_debug) {
+				log_puts("song_strestore: ");
+				ev_log(&s->ev);
+				log_puts(": restored -> ");
+				ev_log(&re);
+				log_puts("\n");
+			}
+			mixout_putev(&re, prio);
+		}
+		s->tag = 1;
 	}
 }
 
@@ -719,30 +938,29 @@ song_confrestore(struct statelist *slist)
  * cancel all frames in the given state list
  */
 void
-song_confcancel(struct statelist *slist)
+song_confcancel(struct statelist *slist, unsigned prio)
 {
 	struct state *s;
 	struct ev ca;
 
-
 	for (s = slist->first; s != NULL; s = s->next) {
 		if (s->tag) {
 			if (state_cancel(s, &ca)) {
 				if (song_debug) {
-					dbg_puts("song_stcancel: ");
-					ev_dbg(&s->ev);
-					dbg_puts(": canceled -> ");
-					ev_dbg(&ca);
-					dbg_puts("\n");
+					log_puts("song_stcancel: ");
+					ev_log(&s->ev);
+					log_puts(": canceled -> ");
+					ev_log(&ca);
+					log_puts("\n");
 				}
-				mixout_putev(&ca, slist->serial);
+				mixout_putev(&ca, prio);
 			}
 			s->tag = 0;
 		} else {
 			if (song_debug) {
-				dbg_puts("song_stcancel: ");
-				ev_dbg(&s->ev);
-				dbg_puts(": not canceled (no tag)\n");
+				log_puts("song_stcancel: ");
+				ev_log(&s->ev);
+				log_puts(": not canceled (no tag)\n");
 			}
 		}
 	}
@@ -755,7 +973,7 @@ void
 song_trkmute(struct song *s, struct songtrk *t)
 {
 	if (s->mode >= SONG_PLAY)
-		song_confcancel(&t->trackptr->statelist);
+		song_confcancel(&t->trackptr->statelist, PRIO_TRACK);
 	t->mute = 1;
 }
 
@@ -766,7 +984,7 @@ void
 song_trkunmute(struct song *s, struct songtrk *t)
 {
 	if (s->mode >= SONG_PLAY)
-		song_confrestore(&t->trackptr->statelist);
+		song_confrestore(&t->trackptr->statelist, 1, PRIO_TRACK);
 	t->mute = 0;
 }
 
@@ -776,25 +994,110 @@ song_trkunmute(struct song *s, struct songtrk *t)
 void
 song_mergerec(struct song *o)
 {
+	struct state *st;
+	struct track loop;
+	struct seqptr *lp;
 	struct songtrk *t;
+	struct songsx *x;
+	struct sysex *e;
 	struct state *s;
 	struct ev ev;
+	unsigned period, offset;
 
 	/*
 	 * if there is no filter for recording there may be
 	 * unterminated frames, so finalize them.
 	 */
-	for (s = o->recptr->statelist.first; s != NULL; s = s->next) {
-		if (!(s->phase & EV_PHASE_LAST) &&
-		     state_cancel(s, &ev)) {
-			seqptr_evput(o->recptr, &ev);
+	for (s = o->rec_input.first; s != NULL; s = s->next) {
+		if (s->tag != TAG_REC && s->tag != TAG_PLAY)
+			continue;
+		if (state_cancel(s, &ev)) {
+			if (s->tag == TAG_REC) {
+				seqptr_evmerge2(o->recptr,
+				    &o->rec_replay, &ev, NULL);
+			}
+			mixout_putev(&ev, 0);
 		}
+		s->tag = TAG_OFF;
 	}
+
+	/*
+	 * cancel any remaining frames (replayed events in loop mode)
+	 */
+	for (s = o->recptr->statelist.first; s != NULL; s = s->next) {
+		if (!state_cancel(s, &ev))
+			continue;
+		mixout_putev(&ev, 0);
+	}
+
+	if (o->playptr != NULL) {
+		period = o->loop_tend - o->loop_tstart;
+		offset = (o->playptr->tic - o->loop_tstart) % period;
+
+		/*
+		 * advance until loop end
+		 */
+		while (offset++ != period) {
+			seqptr_ticdel(o->playptr, 1, &o->rec_replay);
+			seqptr_ticput(o->playptr, 1);
+			seqptr_ticput(o->recptr, 1);
+			for(;;) {
+				st = seqptr_evdel(o->playptr, &o->rec_replay);
+				if (st == NULL)
+					break;
+				st->tag = 1;
+				seqptr_evmerge1(o->recptr, st);
+			}
+		}
+
+		track_init(&loop);
+		lp = seqptr_new(&loop);
+		seqptr_ticput(lp, o->loop_tstart);
+
+		/*
+		 * unroll loop into a new 'loop' track
+		 */
+		while (o->rec.first->ev.cmd != EV_NULL) {
+			seqptr_ticdel(o->playptr, 1, &o->rec_replay);
+			seqptr_ticput(o->playptr, 1);
+			seqptr_ticput(o->recptr, 1);
+			seqptr_ticput(lp, 1);
+			for(;;) {
+				st = seqptr_evdel(o->playptr, &o->rec_replay);
+				if (st == NULL)
+					break;
+				if (st->phase & EV_PHASE_FIRST)
+					st->tag = 0;
+				if (st->tag)
+					seqptr_evmerge1(o->recptr, st);
+				else
+					seqptr_evput(lp, &st->ev);
+			}
+		}
+
+		seqptr_del(lp);
+		track_swap(&o->rec, &loop);
+		track_done(&loop);
+	}
+
 	song_getcurtrk(o, &t);
 	if (t) {
+		undo_track_save(o, &t->track, "record", t->name.str);
 		track_merge(&o->curtrk->track, &o->rec);
+		undo_track_diff(o);
 	}
 	track_clear(&o->rec);
+
+	song_getcursx(o, &x);
+	if (x) {
+		for (;;) {
+			e = sysexlist_get(&o->recsx);
+			if (e == NULL)
+				break;
+			sysexlist_put(&x->sx, e);
+		}
+	}
+	sysexlist_clear(&o->recsx);
 }
 
 /*
@@ -805,12 +1108,13 @@ song_startcb(struct song *o)
 {
 
 	if (song_debug) {
-		dbg_puts("song_startcb:\n");
+		log_puts("song_startcb:\n");
 	}
 	if (o->mode >= SONG_PLAY) {
 		song_ticplay(o);
 		mux_flush();
 	}
+	o->started = 1;
 }
 
 /*
@@ -822,13 +1126,13 @@ song_stopcb(struct song *o)
 	struct songtrk *t;
 
 	if (song_debug)
-		dbg_puts("song_stopcb:\n");
+		log_puts("song_stopcb:\n");
 
 	/*
 	 * stop all sounding notes
 	 */
 	SONG_FOREACH_TRK(o, t) {
-		song_confcancel(&t->trackptr->statelist);
+		song_confcancel(&t->trackptr->statelist, PRIO_TRACK);
 	}
 }
 
@@ -838,15 +1142,6 @@ song_stopcb(struct song *o)
 void
 song_movecb(struct song *o)
 {
-	unsigned delta;
-
-	if (o->mode >= SONG_REC) {
-		while (seqptr_evget(o->recptr))
-			; /* nothing */
-		delta = seqptr_ticskip(o->recptr, 1);
-		if (delta == 0)
-			seqptr_ticput(o->recptr, 1);
-	}
 	if (o->mode >= SONG_PLAY) {
 		(void)song_ticskip(o);
 		song_ticplay(o);
@@ -860,9 +1155,85 @@ song_movecb(struct song *o)
 void
 song_evcb(struct song *o, struct ev *ev)
 {
-	if (o->mode >= SONG_REC) {
-		if (mux_getphase() >= MUX_START)
-			(void)seqptr_evput(o->recptr, ev);
+	struct ev filtout[FILT_MAXNRULES], rev;
+	struct state *s;
+	unsigned i, nev;
+	unsigned usec24;
+
+	if (o->tap_mode != SONG_TAP_OFF &&
+	    evspec_matchev(&o->tap_evspec, ev)) {
+		if (!(ev_phase(ev) & EV_PHASE_FIRST))
+			return;
+		if (o->started)
+			return;
+		if (o->tap_cnt == 0) {
+			if (o->tap_mode == SONG_TAP_START) {
+				log_puts("start triggered\n");
+				o->tap_cnt = -1;
+				mux_ticcb();
+			} else {
+				log_puts("measuring tempo...\n");
+				o->tap_time = mux_wallclock;
+			}
+		} else if (o->tap_mode == SONG_TAP_TEMPO && o->tap_cnt == 1) {
+			usec24 = (long)(mux_wallclock - o->tap_time) / o->tpb;
+			log_puts("start triggered, tempo = ");
+			log_putu(60 * 24000000 / o->tpb / usec24);
+			log_puts("\n");
+			if (usec24 < TEMPO_MIN || usec24 > TEMPO_MAX) {
+				log_puts("tempo out of range, aborted\n");
+				o->tap_cnt = 0;
+				return;
+			}
+			o->tempo = usec24;
+			mux_chgtempo(o->tempo);
+			mux_ticcb();
+			o->tap_cnt = -1;
+		}
+		o->tap_cnt++;
+		return;
+	}
+
+	/*
+	 * apply filter, if any
+	 */
+	if (o->curfilt) {
+		nev = filt_do(&o->curfilt->filt, ev, filtout);
+	} else {
+		filtout[0] = *ev;
+		nev = 1;
+	}
+
+	/*
+	 * output and/or record resulting events
+	 */
+	ev = filtout;
+	for (i = 0; i < nev; i++) {
+		if (o->mode >= SONG_REC) {
+			s = statelist_update(&o->rec_input, ev);
+			if (s->phase & EV_PHASE_FIRST) {
+				s->tic = 0;
+				if (s->flags & (STATE_BOGUS | STATE_NESTED)) {
+					s->tag = TAG_OFF;
+				} else if ((mux_getphase() >= MUX_START) &&
+				    (o->loop_mstart == o->loop_mend ||
+					o->abspos >= o->loop_tstart)) {
+					s->tag = TAG_REC;
+				} else
+					s->tag = TAG_PLAY;
+			}
+
+			if (s->tag == TAG_REC) {
+				if (seqptr_evmerge2(o->recptr,
+					&o->rec_replay, ev, &rev))
+					mixout_putev(&rev, 0);
+			}
+			if (s->tag == TAG_REC || s->tag == TAG_PLAY)
+				mixout_putev(ev, 0);
+
+		} else
+			mixout_putev(ev, 0);
+		ev++;
 	}
 }
 
@@ -873,31 +1244,73 @@ void
 song_sysexcb(struct song *o, struct sysex *sx)
 {
 	if (song_debug)
-		dbg_puts("song_sysexcb:\n");
+		log_puts("song_sysexcb:\n");
 	if (sx == NULL) {
-		dbg_puts("got null sx\n");
+		log_puts("got null sx\n");
 		return;
 	}
-	if (o->mode >= SONG_REC && o->cursx)
-		sysexlist_put(&o->cursx->sx, sx);
+	if (o->mode >= SONG_REC)
+		sysexlist_put(&o->recsx, sx);
 	else
 		sysex_del(sx);
 }
 
+unsigned
+song_mtcpos(struct song *o, unsigned where, unsigned offs)
+{
+	struct seqptr *p;
+	unsigned delta, bpm, tpb, meas, beat, tick;
+	unsigned long long pos;
+	unsigned long usec24;
+
+	p = seqptr_new(&o->meta);
+	pos = 0;
+	meas = beat = tick = 0;
+
+	for (;;) {
+		seqptr_getsign(p, &bpm, &tpb);
+		seqptr_gettempo(p, &usec24);
+		delta = (where - meas) * bpm * tpb - beat * tpb - tick;
+		if (delta <= offs)
+			break;
+		delta -= offs;
+		if (!seqptr_eot(p)) {
+			delta = seqptr_ticskip(p, delta);
+			while (seqptr_evget(p))
+				; /* nothing */
+		}
+		tick += delta;
+		beat += tick / tpb;
+		tick = tick % tpb;
+		meas += beat / bpm;
+		beat = beat % bpm;
+		pos += (unsigned long long)delta * usec24;
+	}
+
+	/* round to frame */
+	pos -= pos % (24000000ULL / DEFAULT_FPS);
+
+	/* wrap every 24 hours */
+	pos = pos % (24000000ULL * 36000 * 24);
+
+	seqptr_del(p);
+	return pos / (24000000ULL / MTC_SEC);
+}
+
 /*
- * cancel the current state, and restore the state of 
+ * cancel the current state, and restore the state of
  * the new postition, must be in idle mode
  *
  * return the absolute position as an ``MTC position'',
  * suitable for a MMC relocate message.
  */
 unsigned
-song_loc(struct song *o, unsigned where, unsigned how)
+song_loc(struct song *o, unsigned how, unsigned where, unsigned offs)
 {
 	struct state *s;
 	struct songtrk *t;
-	unsigned maxdelta, delta, tic;
-	unsigned bpm, tpb, offs;
+	unsigned maxdelta, delta;
+	unsigned bpm, tpb;
 	unsigned long long pos, endpos;
 	unsigned long usec24;
 
@@ -906,48 +1319,46 @@ song_loc(struct song *o, unsigned where, unsigned how)
 
 	/* please gcc */
 	endpos = 0xdeadbeef;
-	offs = 0xdeadbeef;
 	maxdelta = 0xdeadbeef;
 
 	/*
-	 * XXX: when not in LOC_MTC mode, the MTC position
+	 * XXX: when not in LOC_MEAS and LOC_SPP modes, the MTC position
 	 * can overflow. Limit song_loc() to 24 hours
 	 */
 
 	switch (how) {
-	case SONG_LOC_MEAS:
-		offs = (o->mode >= SONG_PLAY) ? o->curquant / 2 : 0;
+	case LOC_MEAS:
 		break;
-	case SONG_LOC_MTC:
+	case LOC_MTC:
 		endpos = (unsigned long long)where * (24000000 / MTC_SEC);
 		offs = 0;
 		break;
-	case SONG_LOC_SPP:
+	case LOC_SPP:
 		where *= o->tics_per_unit / 16;
 		offs = 0;
 		break;
 	default:
-		dbg_puts("song_loc: bad argument\n");
-		dbg_panic();
+		log_puts("song_loc: bad argument\n");
+		panic();
 	}
-	tic = 0;
 	pos = 0;
+	o->abspos = 0;
 	o->measure = o->beat = o->tic = 0;
-		
+
 	for (;;) {
 		seqptr_getsign(o->metaptr, &bpm, &tpb);
 		seqptr_gettempo(o->metaptr, &usec24);
 		switch (how) {
-		case SONG_LOC_MEAS:
+		case LOC_MEAS:
 			maxdelta = (where - o->measure) * bpm * tpb
 			    - o->beat * tpb
 			    - o->tic;
 			break;
-		case SONG_LOC_MTC:
+		case LOC_MTC:
 			maxdelta = (endpos - pos) / usec24;
 			break;
-		case SONG_LOC_SPP:
-			maxdelta = where - tic;
+		case LOC_SPP:
+			maxdelta = where - o->abspos;
 			break;
 		}
 		if (maxdelta <= offs)
@@ -964,11 +1375,21 @@ song_loc(struct song *o, unsigned where, unsigned how)
 		o->beat += o->tic / tpb;
 		o->tic = o->tic % tpb;
 		o->measure += o->beat / bpm;
-		o->beat = o->beat % bpm; 
+		o->beat = o->beat % bpm;
+		o->abspos += delta;
 		pos += (unsigned long long)delta * usec24;
-		tic += delta;
 	}
 
+	/*
+	 * process all meta events of the current tick,
+	 * so we get the time signature and the tempo of
+	 * the next tick
+	 */
+	while (seqptr_evget(o->metaptr))
+		; /* nothing */
+
+	o->complete = !seqptr_eot(o->metaptr);
+
 	/*
 	 * move all tracks to the current position
 	 */
@@ -976,7 +1397,7 @@ song_loc(struct song *o, unsigned where, unsigned how)
 		/*
 		 * cancel and free old states
 		 */
-		song_confcancel(&t->trackptr->statelist);
+		song_confcancel(&t->trackptr->statelist, PRIO_TRACK);
 		statelist_empty(&t->trackptr->statelist);
 		seqptr_del(t->trackptr);
 
@@ -984,57 +1405,92 @@ song_loc(struct song *o, unsigned where, unsigned how)
 		 * allocate and restore new states
 		 */
 		t->trackptr = seqptr_new(&t->track);
-		seqptr_skip(t->trackptr, tic);
-		song_confrestore(&t->trackptr->statelist);
+		seqptr_skip(t->trackptr, o->abspos);
+		for (s = t->trackptr->statelist.first; s != NULL; s = s->next)
+			s->tag = 0;
+		song_confrestore(&t->trackptr->statelist,
+		    o->mode >= SONG_PLAY, PRIO_TRACK);
+
+		/*
+		 * check if we reached the end-of-track
+		 */
+		if (!seqptr_eot(t->trackptr))
+			o->complete = 0;
 	}
 
 	if (o->mode >= SONG_REC)
 		track_clear(&o->rec);
 #ifdef SONG_DEBUG
 	if (!track_isempty(&o->rec)) {
-		dbg_puts("song_loc: rec track not empty\n");
-		dbg_panic();
+		log_puts("song_loc: rec track not empty\n");
+		panic();
 	}
 #endif
+	/*
+	 * at this point track is cleared, song->rec_xxx lists are
+	 * empty
+	 */
+	if (o->playptr)
+		seqptr_del(o->playptr);
 	seqptr_del(o->recptr);
 	o->recptr = seqptr_new(&o->rec);
+	o->playptr = NULL;
 	if (o->mode >= SONG_REC) {
-		seqptr_seek(o->recptr, tic);
+		seqptr_seek(o->recptr, o->abspos);
 	}
+	statelist_empty(&o->rec_input);
+	statelist_empty(&o->rec_replay);
 
+	/*
+	 * we've the tempo to be set, as in the LOC_MTC case, the return
+	 * value is a franction of the tick length. The tempo is set,
+	 * either by mux_open() or by the song_metaput() calls.
+	 */
 	for (s = o->metaptr->statelist.first; s != NULL; s = s->next) {
 		if (EV_ISMETA(&s->ev)) {
 			if (song_debug) {
-				dbg_puts("song_loc: ");
-				ev_dbg(&s->ev);
-				dbg_puts(": restoring meta-event\n");
+				log_puts("song_loc: ");
+				ev_log(&s->ev);
+				log_puts(": restoring meta-event\n");
 			}
 			song_metaput(o, s);
 			s->tag = 1;
 		} else {
 			if (song_debug) {
-				dbg_puts("song_loc: ");
-				ev_dbg(&s->ev);
-				dbg_puts(": not restored (not tagged)\n");
+				log_puts("song_loc: ");
+				ev_log(&s->ev);
+				log_puts(": not restored (not tagged)\n");
 			}
 			s->tag = 0;
 		}
 	}
-	pos /= 24000000 / MTC_SEC;
+
+	if (o->complete)
+		cons_puttag("complete");
+
+	pos = endpos - pos;
+
+	/*
+	 * display initial position
+	 */
+	cons_putpos(o->measure, o->beat, o->tic);
+
 	if (song_debug) {
-		dbg_puts("song_loc: ");
-		dbg_putu(where);
-		dbg_puts(" -> ");
-		dbg_putu(o->measure);
-		dbg_puts(":");
-		dbg_putu(o->beat);
-		dbg_puts(":");
-		dbg_putu(o->tic);
-		dbg_puts("/");
-		dbg_putu(tic);
-		dbg_puts(", mtc = ");
-		dbg_putu(pos);
-		dbg_puts("\n");
+		log_puts("song_loc: ");
+		log_putu(where);
+		log_puts(" -> ");
+		log_putu(o->measure);
+		log_puts(":");
+		log_putu(o->beat);
+		log_puts(":");
+		log_putu(o->tic);
+		log_puts("(");
+		log_putu(o->abspos);
+		log_puts(") +");
+		log_putu(pos);
+		log_puts("/");
+		log_putu(usec24);
+		log_puts("\n");
 	}
 	return pos;
 }
@@ -1047,18 +1503,9 @@ song_loc(struct song *o, unsigned where, unsigned how)
  * this duration to ensure we're perfectly in sync.
  */
 unsigned
-song_gotocb(struct song *o, unsigned mtcpos)
+song_gotocb(struct song *o, int how, unsigned where)
 {
-	unsigned newpos;
-
-	newpos = song_loc(o, mtcpos, SONG_LOC_MTC);
-	cons_putpos(o->measure, o->beat, o->tic);
-
-	if (newpos > mtcpos) {
-		dbg_puts("song_gotocb: negative offset\n");
-		dbg_panic();
-	}
-	return (mtcpos - newpos) * (24000000 / MTC_SEC);
+	return song_loc(o, how, where, 0);
 }
 
 /*
@@ -1067,7 +1514,6 @@ song_gotocb(struct song *o, unsigned mtcpos)
 void
 song_setmode(struct song *o, unsigned newmode)
 {
-	struct songfilt *f;
 	struct songtrk *t;
 	unsigned oldmode;
 
@@ -1080,25 +1526,36 @@ song_setmode(struct song *o, unsigned newmode)
 		metro_setmode(&o->metro, newmode);
 	if (oldmode >= SONG_REC && newmode < SONG_REC)
 		song_mergerec(o);
+	if (oldmode >= SONG_PLAY && newmode < SONG_PLAY)
+		song_loop_done(o);
 	if (oldmode >= SONG_IDLE && newmode < SONG_IDLE) {
 		/*
 		 * cancel and free states
 		 */
 		SONG_FOREACH_TRK(o, t) {
-			song_confcancel(&t->trackptr->statelist);
+			song_confcancel(&t->trackptr->statelist, PRIO_TRACK);
 			statelist_empty(&t->trackptr->statelist);
 			seqptr_del(t->trackptr);
 		}
+		if (o->playptr)
+			seqptr_del(o->playptr);
+		statelist_empty(&o->rec_input);
+		statelist_done(&o->rec_input);
+		statelist_empty(&o->rec_replay);
+		statelist_done(&o->rec_replay);
 		seqptr_del(o->recptr);
 		seqptr_del(o->metaptr);
-		norm_setfilt(NULL);
+		norm_shut();
 		mux_flush();
 		mux_close();
 	}
 	if (oldmode < SONG_PLAY && newmode >= SONG_PLAY) {
+		o->tap_cnt = 0;
 		o->complete = 0;
+		song_loop_init(o);
 	}
 	if (oldmode < SONG_IDLE && newmode >= SONG_IDLE) {
+		o->abspos = 0;
 		o->measure = 0;
 		o->beat = 0;
 		o->tic = 0;
@@ -1111,14 +1568,12 @@ song_setmode(struct song *o, unsigned newmode)
 		}
 		o->metaptr = seqptr_new(&o->meta);
 		o->recptr = seqptr_new(&o->rec);
+		o->playptr = NULL;
+		statelist_init(&o->rec_replay);
+		statelist_init(&o->rec_input);
 
 		mux_open();
-
-		/*
-		 * check for the current filter
-		 */
-		song_getcurfilt(o, &f);
-		norm_setfilt(f != NULL ? &f->filt : NULL);
+		mux_chgticrate(o->tics_per_unit);
 
 		/*
 		 * send sysex messages and channel config messages
@@ -1132,13 +1587,13 @@ song_setmode(struct song *o, unsigned newmode)
 }
 
 /*
- * cancel the current state, and restore the state of 
+ * cancel the current state, and restore the state of
  * the given position
  */
 void
 song_goto(struct song *o, unsigned measure)
 {
-	unsigned mmcpos;
+	unsigned mmcpos, offs;
 
 	if (o->mode >= SONG_IDLE) {
 		/*
@@ -1146,17 +1601,23 @@ song_goto(struct song *o, unsigned measure)
 		 */
 		if (o->mode >= SONG_REC && measure > 0)
 			measure--;
+		o->started = 0;
+
+		offs = (o->mode >= SONG_PLAY && !o->tap_mode) ?
+		    o->curquant / 2 : 0;
 
 		/*
-		 * move all tracks to given measure
+		 * Move all tracks to given measure.
+		 *
+		 * If we're starting playback/recording and using MTC,
+		 * then just send the request, song_loc() will be called
+		 * back later, upon the initial MTC full-frame message.
 		 */
-		mmcpos = song_loc(o, measure, SONG_LOC_MEAS);
-		mux_gotoreq(mmcpos);
-
-		/*
-		 * display initial position
-		 */
-		cons_putpos(o->measure, o->beat, o->tic);
+		if (o->mode >= SONG_PLAY && mididev_mtcsrc != NULL) {
+			mmcpos = song_mtcpos(o, measure, offs);
+			mux_gotoreq(mmcpos);
+		} else
+			song_loc(o, LOC_MEAS, measure, offs);
 	} else
 		cons_putpos(measure, 0, 0);
 }
@@ -1183,10 +1644,11 @@ song_play(struct song *o)
 	m = (o->mode >= SONG_IDLE) ? o->measure : o->curpos;
 	song_setmode(o, SONG_PLAY);
 	song_goto(o, m);
-	mux_startreq();
+	mux_startreq(o->tap_mode != SONG_TAP_OFF);
+	mux_flush();
 
 	if (song_debug) {
-		dbg_puts("song_play: waiting for a start event...\n");
+		log_puts("song_play: waiting for a start event...\n");
 	}
 }
 
@@ -1202,15 +1664,16 @@ song_record(struct song *o)
 
 	song_getcurtrk(o, &t);
 	if (!t || t->mute) {
-		dbg_puts("song_record: no current track (or muted)\n");
+		log_puts("song_record: no current track (or muted)\n");
 	}
 
 	m = (o->mode >= SONG_IDLE) ? o->measure : o->curpos;
 	song_setmode(o, SONG_REC);
 	song_goto(o, m);
-	mux_startreq();
+	mux_startreq(o->tap_mode != SONG_TAP_OFF);
+	mux_flush();
 	if (song_debug) {
-		dbg_puts("song_record: waiting for a start event...\n");
+		log_puts("song_record: waiting for a start event...\n");
 	}
 }
 
@@ -1225,9 +1688,10 @@ song_idle(struct song *o)
 	m = (o->mode >= SONG_IDLE) ? o->measure : o->curpos;
 	song_setmode(o, SONG_IDLE);
 	song_goto(o,  m);
+	mux_flush();
 
 	if (song_debug) {
-		dbg_puts("song_idle: started loop...\n");
+		log_puts("song_idle: started loop...\n");
 	}
 }
 
@@ -1341,18 +1805,46 @@ song_try_meta(struct song *o)
 	return 1;
 }
 
+unsigned
+song_try_ev(struct song *o, unsigned cmd)
+{
+	struct songfilt *f;
+	struct songchan *c;
+	struct songtrk *t;
+
+	SONG_FOREACH_TRK(o, t) {
+		if (track_evcnt(&t->track, cmd)) {
+			cons_errs(t->name.str, "event in use by track");
+			return 0;
+		}
+	}
+	SONG_FOREACH_CHAN(o, c) {
+		if (track_evcnt(&c->conf, cmd)) {
+			cons_errs(c->name.str, "event in use by input");
+			return 0;
+		}
+	}
+	SONG_FOREACH_FILT(o, f) {
+		if (filt_evcnt(&f->filt, cmd)) {
+			cons_errs(f->name.str, "event in use by filter");
+			return 0;
+		}
+	}
+	return 1;
+}
+
 void
-song_confev(struct song *o, struct songchan *c, int input, struct ev *ev)
+song_confev(struct song *o, struct songchan *c, struct ev *ev)
 {
 	track_confev(&c->conf, ev);
 	if (mux_isopen) {
-		song_playconfev(o, c, input, ev);
+		song_playconfev(o, c, ev);
 		mux_flush();
 	}
 }
 
 void
-song_unconfev(struct song *o, struct songchan *c, int input, struct evspec *es)
+song_unconfev(struct song *o, struct songchan *c, struct evspec *es)
 {
 	track_unconfev(&c->conf, es);
 }
